module StringMap = Map.Make (String)

type error =
  | Illegal_character of char
  | Unterminated_string
  | Unterminated_comment
  | Illegal_escape of string
  | Unexpected_token
  | Expect_comma_or_rbracket
  | Expect_comma_or_rbrace
  | Expect_colon
  | Expect_string_or_rbrace
  | Expect_eof
(* | Trailing_comma_in_obj *)
(* | Trailing_comma_in_array *)

let fprintf = Format.fprintf

let report_error ppf = function
  | Illegal_character c -> fprintf ppf "Illegal character (%s)" (Char.escaped c)
  | Illegal_escape s ->
    fprintf ppf "Illegal backslash escape in string or character (%s)" s
  | Unterminated_string -> fprintf ppf "Unterminated_string"
  | Expect_comma_or_rbracket -> fprintf ppf "Expect_comma_or_rbracket"
  | Expect_comma_or_rbrace -> fprintf ppf "Expect_comma_or_rbrace"
  | Expect_colon -> fprintf ppf "Expect_colon"
  | Expect_string_or_rbrace -> fprintf ppf "Expect_string_or_rbrace"
  | Expect_eof -> fprintf ppf "Expect_eof"
  | Unexpected_token -> fprintf ppf "Unexpected_token"
  (* | Trailing_comma_in_obj  *)
  (*   -> fprintf ppf "Trailing_comma_in_obj" *)
  (* | Trailing_comma_in_array  *)
  (*   -> fprintf ppf "Trailing_comma_in_array" *)
  | Unterminated_comment -> fprintf ppf "Unterminated_comment"

exception Error of Lexing.position * Lexing.position * error

let () =
  Printexc.register_printer (function x ->
      (match x with
      | Error (loc_start, loc_end, error) ->
        Some
          (Format.asprintf "@[%a:@ %a@ -@ %a)@]" report_error error
             Ext_position.print loc_start Ext_position.print loc_end)
      | _ -> None))

type token =
  | Comma
  | Eof
  | False
  | Lbrace
  | Lbracket
  | Null
  | Colon
  | Number of string
  | Rbrace
  | Rbracket
  | String of string
  | True

let error (lexbuf : Lexing.lexbuf) e =
  raise (Error (lexbuf.lex_start_p, lexbuf.lex_curr_p, e))
  [@@raises Error]

let lexeme_len (x : Lexing.lexbuf) = x.lex_curr_pos - x.lex_start_pos

let update_loc ({lex_curr_p; _} as lexbuf : Lexing.lexbuf) diff =
  lexbuf.lex_curr_p <-
    {
      lex_curr_p with
      pos_lnum = lex_curr_p.pos_lnum + 1;
      pos_bol = lex_curr_p.pos_cnum - diff;
    }

let char_for_backslash = function
  | 'n' -> '\010'
  | 'r' -> '\013'
  | 'b' -> '\008'
  | 't' -> '\009'
  | c -> c

let dec_code c1 c2 c3 =
  (100 * (Char.code c1 - 48)) + (10 * (Char.code c2 - 48)) + (Char.code c3 - 48)

let hex_code c1 c2 =
  let d1 = Char.code c1 in
  let val1 =
    if d1 >= 97 then d1 - 87 else if d1 >= 65 then d1 - 55 else d1 - 48
  in
  let d2 = Char.code c2 in
  let val2 =
    if d2 >= 97 then d2 - 87 else if d2 >= 65 then d2 - 55 else d2 - 48
  in
  (val1 * 16) + val2

let lf = '\010'

let __ocaml_lex_tables =
  {
    Lexing.lex_base =
      "\000\000\239\255\240\255\241\255\000\000\025\000\011\000\244\255\245\255\246\255\247\255\248\255\249\255\000\000\000\000\000\000\041\000\001\000\254\255\005\000\005\000\253\255\001\000\002\000\252\255\000\000\000\000\003\000\251\255\001\000\003\000\250\255\079\000\089\000\099\000\121\000\131\000\141\000\153\000\163\000\001\000\253\255\254\255\023\000\255\255\006\000\246\255\189\000\248\255\215\000\255\255\249\255\249\000\181\000\252\255\009\000\063\000\075\000\234\000\251\255\032\001\250\255";
    Lexing.lex_backtrk =
      "\255\255\255\255\255\255\255\255\013\000\013\000\016\000\255\255\255\255\255\255\255\255\255\255\255\255\016\000\016\000\016\000\016\000\016\000\255\255\000\000\012\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\013\000\255\255\013\000\255\255\013\000\255\255\255\255\255\255\255\255\001\000\255\255\255\255\255\255\008\000\255\255\255\255\255\255\255\255\006\000\006\000\255\255\006\000\001\000\002\000\255\255\255\255\255\255\255\255";
    Lexing.lex_default =
      "\001\000\000\000\000\000\000\000\255\255\255\255\255\255\000\000\000\000\000\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\000\000\255\255\020\000\000\000\255\255\255\255\000\000\255\255\255\255\255\255\000\000\255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\042\000\000\000\000\000\255\255\000\000\047\000\000\000\047\000\000\000\051\000\000\000\000\000\255\255\255\255\000\000\255\255\255\255\255\255\255\255\000\000\255\255\000\000";
    Lexing.lex_trans =
      "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\019\000\018\000\018\000\019\000\017\000\019\000\255\255\048\000\019\000\255\255\057\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\019\000\000\000\003\000\000\000\000\000\019\000\000\000\000\000\050\000\000\000\000\000\043\000\008\000\006\000\033\000\016\000\004\000\005\000\005\000\005\000\005\000\005\000\005\000\005\000\005\000\005\000\007\000\004\000\005\000\005\000\005\000\005\000\005\000\005\000\005\000\005\000\005\000\032\000\044\000\033\000\056\000\005\000\005\000\005\000\005\000\005\000\005\000\005\000\005\000\005\000\005\000\021\000\057\000\000\000\000\000\000\000\020\000\000\000\000\000\012\000\000\000\011\000\032\000\056\000\000\000\025\000\049\000\000\000\000\000\032\000\014\000\024\000\028\000\000\000\000\000\057\000\026\000\030\000\013\000\031\000\000\000\000\000\022\000\027\000\015\000\029\000\023\000\000\000\000\000\000\000\039\000\010\000\039\000\009\000\032\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\037\000\000\000\037\000\000\000\035\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\255\255\035\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\000\000\000\000\255\255\000\000\056\000\000\000\000\000\055\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\054\000\000\000\054\000\000\000\000\000\000\000\000\000\054\000\000\000\002\000\041\000\000\000\000\000\000\000\255\255\046\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\255\255\059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\000\000\000\000\000\000\000\000\000\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\054\000\000\000\000\000\000\000\000\000\000\000\054\000\060\000\060\000\060\000\060\000\060\000\060\000\000\000\000\000\000\000\000\000\000\000\054\000\000\000\000\000\000\000\054\000\000\000\054\000\000\000\000\000\000\000\052\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\060\000\060\000\060\000\060\000\060\000\060\000\000\000\061\000\061\000\061\000\061\000\061\000\061\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\061\000\061\000\061\000\061\000\061\000\061\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000";
    Lexing.lex_check =
      "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\000\000\017\000\000\000\000\000\019\000\020\000\045\000\019\000\020\000\055\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\000\000\255\255\255\255\019\000\255\255\255\255\045\000\255\255\255\255\040\000\000\000\000\000\004\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\006\000\006\000\006\000\006\000\006\000\006\000\006\000\006\000\006\000\006\000\004\000\043\000\005\000\056\000\005\000\005\000\005\000\005\000\005\000\005\000\005\000\005\000\005\000\005\000\016\000\057\000\255\255\255\255\255\255\016\000\255\255\255\255\000\000\255\255\000\000\005\000\056\000\255\255\014\000\045\000\255\255\255\255\004\000\000\000\023\000\027\000\255\255\255\255\057\000\025\000\029\000\000\000\030\000\255\255\255\255\015\000\026\000\000\000\013\000\022\000\255\255\255\255\255\255\032\000\000\000\032\000\000\000\005\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\035\000\255\255\035\000\255\255\034\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\047\000\034\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\039\000\039\000\039\000\039\000\039\000\039\000\039\000\039\000\039\000\039\000\255\255\255\255\047\000\255\255\049\000\255\255\255\255\049\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\049\000\255\255\049\000\255\255\255\255\255\255\255\255\049\000\255\255\000\000\040\000\255\255\255\255\255\255\020\000\045\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\047\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\255\255\255\255\255\255\255\255\255\255\052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\049\000\255\255\255\255\255\255\255\255\255\255\049\000\052\000\052\000\052\000\052\000\052\000\052\000\255\255\255\255\255\255\255\255\255\255\049\000\255\255\255\255\255\255\049\000\255\255\049\000\255\255\255\255\255\255\049\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\052\000\052\000\052\000\052\000\052\000\052\000\255\255\060\000\060\000\060\000\060\000\060\000\060\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\060\000\060\000\060\000\060\000\060\000\060\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\047\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\049\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255";
    Lexing.lex_base_code = "";
    Lexing.lex_backtrk_code = "";
    Lexing.lex_default_code = "";
    Lexing.lex_trans_code = "";
    Lexing.lex_check_code = "";
    Lexing.lex_code = "";
  }

let rec lex_json buf lexbuf = __ocaml_lex_lex_json_rec buf lexbuf 0
  [@@raises Error]

and __ocaml_lex_lex_json_rec buf lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
  | 0 -> lex_json buf lexbuf
  | 1 ->
    update_loc lexbuf 0;
    lex_json buf lexbuf
  | 2 -> comment buf lexbuf
  | 3 -> True
  | 4 -> False
  | 5 -> Null
  | 6 -> Lbracket
  | 7 -> Rbracket
  | 8 -> Lbrace
  | 9 -> Rbrace
  | 10 -> Comma
  | 11 -> Colon
  | 12 -> lex_json buf lexbuf
  | 13 -> Number (Lexing.lexeme lexbuf)
  | 14 ->
    let pos = Lexing.lexeme_start_p lexbuf in
    scan_string buf pos lexbuf;
    let content = Buffer.contents buf in
    Buffer.clear buf;
    String content
  | 15 -> Eof
  | 16 ->
    let c = Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
    error lexbuf (Illegal_character c)
  | __ocaml_lex_state ->
    lexbuf.Lexing.refill_buff lexbuf;
    __ocaml_lex_lex_json_rec buf lexbuf __ocaml_lex_state
  [@@raises Error]

and comment buf lexbuf = __ocaml_lex_comment_rec buf lexbuf 40 [@@raises Error]

and __ocaml_lex_comment_rec buf lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
  | 0 -> lex_json buf lexbuf
  | 1 -> comment buf lexbuf
  | 2 -> error lexbuf Unterminated_comment
  | __ocaml_lex_state ->
    lexbuf.Lexing.refill_buff lexbuf;
    __ocaml_lex_comment_rec buf lexbuf __ocaml_lex_state
  [@@raises Error]

and scan_string buf start lexbuf =
  __ocaml_lex_scan_string_rec buf start lexbuf 45
  [@@raises Error]

and __ocaml_lex_scan_string_rec buf start lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
  | 0 -> ()
  | 1 ->
    let len = lexeme_len lexbuf - 2 in
    update_loc lexbuf len;

    scan_string buf start lexbuf
  | 2 ->
    let len = lexeme_len lexbuf - 3 in
    update_loc lexbuf len;
    scan_string buf start lexbuf
  | 3 ->
    let c = Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 1) in
    Buffer.add_char buf (char_for_backslash c);
    scan_string buf start lexbuf
  | 4 ->
    let c1 = Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 1)
    and c2 = Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 2)
    and c3 = Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 3)
    and s =
      Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos
        (lexbuf.Lexing.lex_start_pos + 4)
    in
    let v = dec_code c1 c2 c3 in
    (try Buffer.add_char buf (Char.chr v)
     with _ -> error lexbuf (Illegal_escape s));
    scan_string buf start lexbuf
  | 5 ->
    let c1 = Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 2)
    and c2 = Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 3) in
    let v = hex_code c1 c2 in
    (try Buffer.add_char buf (Char.chr v)
     with _ -> error lexbuf (Illegal_escape (Char.escaped c2)));

    scan_string buf start lexbuf
  | 6 ->
    let c = Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 1) in
    Buffer.add_char buf '\\';
    Buffer.add_char buf c;

    scan_string buf start lexbuf
  | 7 ->
    update_loc lexbuf 0;
    Buffer.add_char buf lf;

    scan_string buf start lexbuf
  | 8 ->
    let ofs = lexbuf.lex_start_pos in
    let len = lexbuf.lex_curr_pos - ofs in
    Buffer.add_substring buf (Bytes.to_string lexbuf.lex_buffer) ofs len;

    scan_string buf start lexbuf
  | 9 -> error lexbuf Unterminated_string
  | __ocaml_lex_state ->
    lexbuf.Lexing.refill_buff lexbuf;
    __ocaml_lex_scan_string_rec buf start lexbuf __ocaml_lex_state
  [@@raises Error]

let parse_json lexbuf =
  let buf = Buffer.create 64 in
  let look_ahead = ref None in
  let token () : token =
    match !look_ahead with
    | None -> lex_json buf lexbuf
    | Some x ->
      look_ahead := None;
      x
    [@@raises Error]
  in
  let push e = look_ahead := Some e in
  let rec json (lexbuf : Lexing.lexbuf) : Ext_json_types.t =
    match token () with
    | True -> True lexbuf.lex_start_p
    | False -> False lexbuf.lex_start_p
    | Null -> Null lexbuf.lex_start_p
    | Number s -> Flo s
    | String s -> Str s
    | Lbracket -> parse_array lexbuf.lex_start_p lexbuf.lex_curr_p [] lexbuf
    | Lbrace -> parse_map lexbuf.lex_start_p StringMap.empty lexbuf
    | _ -> error lexbuf Unexpected_token
    (* Note if we remove [trailing_comma] support
        we should report errors (actually more work), for example
        {[
        match token () with
        | Rbracket ->
          if trailing_comma then
            error lexbuf Trailing_comma_in_array
          else
        ]}
        {[
        match token () with
        | Rbrace ->
          if trailing_comma then
            error lexbuf Trailing_comma_in_obj
          else

        ]}
    *)
    [@@raises Error]
  and parse_array loc_start loc_finish acc lexbuf : Ext_json_types.t =
    match token () with
    | Rbracket -> Arr (Array.of_list (acc |> List.rev))
    | x -> (
      push x;
      let new_one = json lexbuf in
      match token () with
      | Comma -> parse_array loc_start loc_finish (new_one :: acc) lexbuf
      | Rbracket -> Arr (Array.of_list (new_one :: acc |> List.rev))
      | _ -> error lexbuf Expect_comma_or_rbracket)
    [@@raises Error]
  and parse_map loc_start acc lexbuf : Ext_json_types.t =
    match token () with
    | Rbrace -> Obj acc
    | String key -> (
      match token () with
      | Colon -> (
        let value = json lexbuf in
        match token () with
        | Rbrace -> Obj (StringMap.add key value acc)
        | Comma -> parse_map loc_start (StringMap.add key value acc) lexbuf
        | _ -> error lexbuf Expect_comma_or_rbrace)
      | _ -> error lexbuf Expect_colon)
    | _ -> error lexbuf Expect_string_or_rbrace
    [@@raises Error]
  in

  let v = json lexbuf in
  match token () with
  | Eof -> v
  | _ -> error lexbuf Expect_eof
  [@@raises Error]

let parse_json_from_file s =
  let in_chan = open_in s in
  match
    let lexbuf = Ext_position.lexbuf_from_channel_with_fname in_chan s in
    parse_json lexbuf
  with
  | exception (Error _ | Invalid_argument _ | Sys_error _) ->
    close_in_noerr in_chan;
    None
  | v ->
    close_in_noerr in_chan;
    Some v
